#!/usr/bin/env tclsh

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.

set binpath [file dirname $argv0]
set dir [file normalize $binpath]/lib

package require getopt
package require uriencode
package require can2svg

set VERSION 0.4
set COPYRIGHT "Copyright (C) 2021-2023 Claude Rubinson"
set NAME Fiss
set DESCR "Fiss Chart Generator"
set DETAIL "Produce publication-quality Fiss configuration charts."


# defaults
set myfont Helvetica
set myfontsize 12
set myheader Recipes
set output_eps 0

# commandline processing
getopt arg $argv {
    -y: - --input:FILE {
	# when FILE is -, read standard input
	if {[string trim $arg] == {-}} {
	    set infile stdin
	} {
	    set infile $arg
	}
    }
    -f: - --font:FONT {
	# font family (default is Helvetica)
	set myfont $arg
    }
    -s: - --size:SIZE {
	# font size (default is 12)
	set myfontsize $arg
    }
    -h: - --header:HEADER {
	# column header (default is "Recipes")
	set myheader $arg
    }
    --bellcurve {
	# dump example config file
	puts [read [open $binpath/bellcurve.fiss r]]
	exit 0
    }
    --greckhamer {
	# dump example config file
	puts [read [open $binpath/greckhamer.fiss r]]
	exit 0
    }
    --template {
	# dump template config file
	puts [read [open $binpath/template.fiss r]]
	exit 0
    }
    -h? - --help {
	# display this help and exit
	help
	exit 0
    }
    --version {
	# output version information and exit
	global VERSION
	global COPYRIGHT
	puts "fisschart $VERSION\n$COPYRIGHT"
	exit 0
    }
}

# need to delay loading of Tk to avoid commandline argument errors
package require Tk

# create canvas
tk::canvas .c -height 600 -width 400 -yscrollcommand {.sy set} -xscrollcommand {.sx set} -bg white

# scrollbars
scrollbar .sx -command ".c xview" -orient horizontal
scrollbar .sy -command ".c yview" -orient vertical

# main menu
option add *tearOff 0
menu .menubar
. configure -menu .menubar
set m .menubar
$m add cascade -menu $m.file -label File -underline 0
$m add cascade -menu $m.help -label Help

# file menu
menu $m.file

$m.file add command -label "Open" -underline 0 -command {
    set f [tk_getOpenFile]
    if {[string equal $f ""] == 0} {
	set ch [open $f r]
	renderChart $ch $f
    }
}

$m.file add command -label "Export as Postscript" -underline 0 -command {
    set f [tk_getSaveFile -initialfile fiss.eps]
    if {[string equal $f ""] == 0} {
	foreach {x0 y0 x1 y1} [.c bbox all] {}
	set w [expr $x1-$x0]
	set h [expr $y1-$y0]
	set ch [open $f w]
	puts $ch [.c postscript -x $x0 -y $y0 -width $w -height $h]
	close $ch
    }
}

$m.file add command -label "Export as SVG" -underline 0 -state disabled -command {
    set f [tk_getSaveFile -initialfile fiss.svg]
    if {[string equal $f ""] == 0} {
	can2svg::canvas2file .c $f
	# foreach {x0 y0 x1 y1} [.c bbox all] {}
	# set w [expr $x1-$x0]
	# set h [expr $y1-$y0]
	# set ch [open $f w]
	# puts $ch [.c postscript -x $x0 -y $y0 -width $w -height $h]
	# close $ch
    }
}

$m.file add separator
$m.file add command -label "Quit" -underline 0 -command {destroy .}

# help menu
menu $m.help
$m.help add command -label "Docs" -command {}

# should probably create/find a reusable widget for the About dialog
$m.help add command -label "About" -command {about}

# grid the widgets
grid .c -sticky nwes -column 0 -row 0
grid columnconfigure . 0 -weight 1
grid rowconfigure . 0 -weight 1
grid .sx -column 0 -row 1 -sticky ew 
grid .sy -column 1 -row 0 -sticky ns

font create MyDefaultFont -family $myfont -size $myfontsize
font create MyHeadingFont -family $myfont -size $myfontsize -weight bold

proc about [] {
    global VERSION
    global COPYRIGHT
    global NAME
    global DESCR
    global DETAIL
    tk_messageBox -title "About $NAME" -message "$DESCR" -detail "$DETAIL\n\nVersion $VERSION\n$COPYRIGHT"
}

proc renderChart {inchan fname} {

    .c delete all

    set mydata [read $inchan]
    
    # titlebar
    if {[dict exists $mydata title] &&
	[string length [dict get $mydata title]]} {
	wm title . "Fiss Chart - $fname ([dict get $mydata title])"
    } {
	wm title . "Fiss Chart - $fname"
    }    

    # scale everything based off of the width of the column headers; note
    # that we don't scale based on font height.  line height is fixed at
    # 20 points, which generally requires a font size of 10 or 12 points
    # to look good. (see below warning about ypos variable)
    #
    # the column width should be the width of 0.99 (the largest possible
    # column value) or the size of the widest column header, whichever is
    # larger, plus 10 points of padding.
    set col_headers [dict get $mydata cols]
    global col_width ;# this should probably be moved into an
		      # application state array of some sort? and/or
		      # we just pass the variable explicitly into each
		      # function.
    set col_width [expr {10 + [font measure MyDefaultFont 0.99]}]
    set header_width [expr {10 + [tcl::mathfunc::max {*}[lmap x $col_headers {font measure MyDefaultFont $x}]]}]
    set col_header_width [tcl::mathfunc::max $col_width $header_width]

    # the below procs create the relevant glyphs; there's a separate proc
    # for each glyph, which is kind of stupid because they're pretty
    # similar to one another, but it's straightforward to update them.
    #
    # the magic is in setting the adjustment and offset values.  adjust
    # scales the glyph according to the column data width (not column
    # header width).  offset allows the glyphs to be set at the origin
    # (rather than upper-left, which is the standard approach).  the
    # necessity glyphs are a bit different, because we're just drawing
    # diamonds.
    #
    # the glyphs appear to be printed a touch too low.  i've added a
    # slight additional vertical offset below to compensate.  very
    # hacky.
    proc corepos {canvas x y} {
	global col_width
	set y [expr {$y - 2}]
	set adjust [expr {$col_width * 0.25}]
	set offset [expr {$adjust / 2}]
	set x [expr $x - $offset]
	set y [expr $y - $offset]
	$canvas create oval $x $y [expr {$x + $adjust}] [expr {$y + $adjust}] -fill black
    }

    proc coreneg {canvas x y} {
	global col_width
	set y [expr {$y - 2}]
	set adjust [expr {$col_width * 0.25}]
	set offset [expr {$adjust / 2}]
	set x [expr $x - $offset]
	set y [expr $y - $offset]
	set x2 [expr {$x+$adjust}]
	set y2 [expr {$y+$adjust}]
	$canvas create oval $x $y [expr {$x + $adjust}] [expr {$y + $adjust}] -width 0.5
	$canvas create line $x [expr {($y+$y2)/2}] $x2 [expr {($y+$y2)/2}]
    }

    proc contribpos {canvas x y} {
	global col_width
	set y [expr {$y - 1}]
	set adjust [expr {$col_width * 0.14}]
	set offset [expr {$adjust / 2}]
	set x [expr $x - $offset]
	set y [expr $y - $offset]
	$canvas create oval $x $y [expr {$x + $adjust}] [expr {$y + $adjust}] -fill black
    }

    proc contribneg {canvas x y} {
	global col_width
	set y [expr {$y - 1}]
	set adjust [expr {$col_width * 0.14}]
	set offset [expr {$adjust / 2}]
	set x [expr $x - $offset]
	set y [expr $y - $offset]
	set x2 [expr {$x+$adjust}]
	set y2 [expr {$y+$adjust}]
	$canvas create oval $x $y [expr {$x + $adjust}] [expr {$y + $adjust}] -width 0.5
	$canvas create line $x [expr {($y+$y2)/2}] $x2 [expr {($y+$y2)/2}]
    }

    proc necpos {canvas x y} {
	global col_width
	set y [expr {$y - 2}]
	set adjust [expr {$col_width * 0.16}]
	set x1 $x
	set y1 [expr {$y-$adjust}]
	set x2 [expr {$x-$adjust}]
	set y2 $y
	set x3 $x
	set y3 [expr {$y+$adjust}]
	set x4 [expr {$x+$adjust}]
	set y4 $y
	$canvas create polygon $x1 $y1 $x2 $y2 $x3 $y3 $x4 $y4
    }

    proc necneg {canvas x y} {
	global col_width
	set y [expr {$y - 2}]
	set adjust [expr {$col_width * 0.14}]
	set x1 $x
	set y1 [expr {$y-$adjust}]
	set x2 [expr {$x-$adjust}]
	set y2 $y
	set x3 $x
	set y3 [expr {$y+$adjust}]
	set x4 [expr {$x+$adjust}]
	set y4 $y
	$canvas create line $x1 $y1 $x2 $y2 $x3 $y3 $x4 $y4 $x1 $y1
    }

    # no-op
    proc irr {canvas x y} {}
    
    # one thing I've thought a lot about is permitting the user to rename
    # the conditions (and recipes) as part of the config file.  but the
    # current implementation doesn't make that very easy to implement.
    # probably the cleanest way to do this would to extend the stub entry
    # to be a list of pairs, the name of the condition and then the name
    # to rewrite (or it could be blank if it didn't need to be rewritten).
    # but I'll leave it out for now until I see how much I really want it.
    set row_headers "[dict get $mydata stub] Consistency {Raw Coverage} {Unique Coverage}"

    set stub_width [expr {25 + [tcl::mathfunc::max {*}[lmap x $row_headers {font measure MyDefaultFont $x}]]}]

    # head
    set xpos $stub_width
    foreach col [dict get $mydata cols] {
	# global xpos ; # an error?
	# global col_header_width
	if {$col != "vrule" && $col != "vrule_dash"} {
	    # stuff all of the x positions into an array for later use
	    set xpositions($col) $xpos
	    .c create text $xpos 30 -text $col -font MyDefaultFont
	    incr xpos $col_header_width
	}
    }
    # $table_width is calculated to extend a bit beyound the final column
    # of data, for aesthetics
    set table_width [expr {$xpos - ($col_header_width/2)}]
    set colhead_midpoint [expr {($stub_width + $xpos - $col_header_width)/2}]
    global myheader
    .c create text $colhead_midpoint 10 -text $myheader -font MyDefaultFont

    # midrule
    .c create line 0 40 $table_width 40

    # stub

    # be careful with ypos, which is initially instantiated below.  it
    # tracks state and is incr'd as the chart is built downward.
    # moreover, it's incr'd in multiples of 20 points, which is the magic
    # value that I'm using for a vskip.  might consider scaling this
    # according the point size of the specified font, but that's a
    # difficult to pull off correctly.  need to think about it some more.

    # a header is indicated by prefixing it with a plus sign; the literal
    # "vskip" adds a bit of blank space to separate groups of conditions
    #
    # we start at the xpositions at 5 to provide a little bit of indent.
    # there should be a way to calculate the proper amount of indent based
    # off of $table_width but I'm not sure how to do that offhand.
    set stub [dict get $mydata stub]
    set ypos 60

    foreach rhead $stub {
	# stuff all of the y values into an array for later use
	set ypositions($rhead) $ypos
	if {$rhead == "vskip"} {
	    incr ypos 10
	} {
	    if {[string compare -length 1 $rhead "+"] == 0} {
		# add extra space before subheader, but not for the
		# first subheader
		if {$ypos > 60} {incr ypos 10}
		.c create text 5 $ypos -text [string replace $rhead 0 0] -anchor w -font MyHeadingFont
	    } {
		.c create text 5 $ypos -text $rhead -anchor w -font MyDefaultFont
	    }
	    if {$rhead == "vskip"} {
		incr ypos 10
	    } {
		incr ypos 20
	    }
	}
    }

    .c create text 5 [incr ypos 20] -text "Consistency" -anchor w -font MyDefaultFont
    set ypositions(con) $ypos
    .c create text 5 [incr ypos 20] -text "Raw coverage" -anchor w -font MyDefaultFont
    set ypositions(cov) $ypos
    .c create text 5 [incr ypos 20] -text "Unique coverage" -anchor w -font MyDefaultFont
    set ypositions(ucov) $ypos

    # bottom rule
    set table_height [incr ypos 20]
    .c create line 0 $table_height $table_width $table_height

    set solcon [dict get $mydata solcon]
    set solcov [dict get $mydata solcov]
    .c create text 5 [incr ypos 20] -text "Solution consistency: [format %1.2f $solcon]" -anchor w -font MyDefaultFont
    .c create text 5 [incr ypos 20] -text "Solution coverage: [format %1.2f $solcov]" -anchor w -font MyDefaultFont

    # vrules
    #
    # there are two style of vrules.  "vrule" is the standard solid line.
    # "vrule_dash" is dashed, as a way of indicating subdivisions within
    # groups of recipes.
    set xpos $stub_width
    foreach col [dict get $mydata cols] {
	# global xpos ; # an error?
	# global col_header_width
	set vrule_pos [expr {($xpos-($col_header_width/2))}]
	set vrule_height [expr {$table_height - 5}]
	if {[string compare $col "vrule"] == 0} {
	    .c create line $vrule_pos 45 $vrule_pos $vrule_height
	} elseif {[string compare $col "vrule_dash"] == 0} {
	    for {set y 45} {$y < $vrule_height} {incr y 10} {
		.c create line $vrule_pos $y $vrule_pos [expr {$y+5}]
	    }
	} else {
	    incr xpos $col_header_width
	}
    }

    # error checking
    #
    # warning: the next few stanzas build on one another; be careful
    # with refactoring!

    # check that every column header corresponds to a recipe name
    set recipe_names [dict keys [dict get $mydata recipes]]
    foreach col [dict get $mydata cols] {
	if {[string compare $col "vrule"] || [string compare $col "vrule_dash"]} {
	    continue
	} else {
	    if {[lsearch -exact $recipe_names $col] == -1} {
		puts stderr "recipe not found: $col"
		exit 1
	    }
	}
    }

    # check that every condition in a recipe is present in the stub
    foreach recipe $recipe_names {
	foreach cond [dict get $mydata recipes $recipe conds] {
	    lappend conds [lindex $cond 0]
	}
    }
    set conds [lsort -unique $conds]
    foreach cond [lsort -unique $conds] {
	if {[lsearch -exact [dict get $mydata stub] $cond] == -1} {
	    puts stderr "condition not found in stub: $cond"
	    exit 1
	}
    }

    # recipes
    #
    # need to put a try/catch here in case any of the dict values are
    # missing or named incorrectly.
    set recipes [dict get $mydata recipes]
    foreach recipe [dict get $mydata cols] {
	if {$recipe != "vrule" && $recipe != "vrule_dash"} {
	    foreach cond [dict get $recipes $recipe conds] {
		# construct and execute glyph output command
		eval [string tolower [lindex $cond 1]] \
		    .c \
		    $xpositions($recipe) \
		    $ypositions([lindex $cond 0])
	    }
	    set con_formatted [format %1.2f [dict get $recipes $recipe con]]
	    set cov_formatted [format %1.2f [dict get $recipes $recipe cov]]
	    set ucov_formatted [format %1.2f [dict get $recipes $recipe ucov]]
	    .c create text $xpositions($recipe) $ypositions(con) -text $con_formatted -font MyDefaultFont
	    .c create text $xpositions($recipe) $ypositions(cov) -text $cov_formatted -font MyDefaultFont
	    .c create text $xpositions($recipe) $ypositions(ucov) -text $ucov_formatted -font MyDefaultFont
}
    }

    #
    # key
    #
    incr ypos 10

    # only include if there's a necessary condition in the chart
    set xpos 20
    if {[string match -nocase *Necpos* $mydata]} {
	necpos .c $xpos [incr ypos 20]
	.c create text [incr xpos 12] $ypos -text "/" -font MyDefaultFont
	necneg .c [incr xpos 12] $ypos
	.c create text [incr xpos 20] $ypos -text "Presence/absence of condition is necessary" -anchor w -font MyDefaultFont
    }

    # use different sufficiency keys, depending upon if both core and
    # contributory conditions are present
    if {[string match -nocase *Core* $mydata] &&
	[string match -nocase *Contrib* $mydata]} then {

	set xpos 20
	corepos .c $xpos [incr ypos 20]
	.c create text [incr xpos 12] $ypos -text "/" -font MyDefaultFont
	contribpos .c [incr xpos 12] $ypos
	.c create text [incr xpos 20] $ypos -text "Core/contributory condition present" -anchor w -font MyDefaultFont

	set xpos 20
	coreneg .c $xpos [incr ypos 20]
	.c create text [incr xpos 12] $ypos -text "/" -font MyDefaultFont
	contribneg .c [incr xpos 12] $ypos
	.c create text [incr xpos 20] $ypos -text "Core/contributory condition absent" -anchor w -font MyDefaultFont

    } else {

	# use this if not distinguishing between core/contributory conditions
	set xpos 20
	corepos .c $xpos [incr ypos 20]
	.c create text [incr xpos 12] $ypos -text "/" -font MyDefaultFont
	coreneg .c [incr xpos 12] $ypos
	.c create text [incr xpos 20] $ypos -text "Explanatory condition is present/absent" -anchor w -font MyDefaultFont
    }

    update
    
}

#
# if reading input from commandline, render image immediately
#
if {[info exists infile]} {
    renderChart stdin stdin
}

# define scroll region, now that we know it
.c configure -scrollregion [.c bbox all]
